import numpy as np

from rock_physics_open.equinor_utilities.various_utilities.types import Array1D, Array2D

from .t_matrix_C import t_matrix_porosity_c_alpha_v


def curve_fit_2_inclusion_sets(
    x_data: Array2D[np.float64],
    frac_ani: float,
    frac_con: float,
    alpha1: float,
    alpha2: float,
    v1: float,
) -> Array1D[np.float64]:
    """Optimisation of input parameters to T-Matrix for carbonate in case where the mineral composition for each
    sample is known, so that effective mineral moduli for each sample are part of the inputs. The optimisation
    is made for the inclusion parameters in addition to fraction of connected and anisotropic porosity and the
    VTI plane angle.

    Parameters
    ----------
    x_data : np.ndarray
        Inputs to unpack.
    frac_ani : float
        Fraction of anisotropic inclusions.
    frac_con : float
        Fraction of connected inclusions.
    alpha1 : float
        Aspect ratio of first inclusion set.
    alpha2 : float
        Aspect ratio of second inclusion set.
    v1 : float
        Concentration ratio of first inclusion set.

    Returns
    -------
    np.ndarray
        Modelled velocity.
    """
    # Unpack x inputs
    # In calling function:
    # x_data = np.stack((por, k_min, mu_min, rho_min, k_fl, rho_fl, k_r, eta_f, tau, freq, def_vp_vs_ratio), axis=1)
    phi = x_data[:, 0]
    k_min = x_data[:, 1]
    mu_min = x_data[:, 2]
    rho_min = x_data[:, 3]
    k_fl = x_data[:, 4]
    rho_fl = x_data[:, 5]
    angle_sym_plane = x_data[0, 6]
    perm = x_data[:, 7]
    visco = x_data[:, 8]
    tau = x_data[:, 9]
    # Both frequency and vp/vs ratio has to be in the same format as the others, but only a scalar value is used
    freq = x_data[0, 10]
    def_vp_vs_ratio = x_data[0, 11]

    log_length = len(phi)

    # alpha and v values should follow these conditions:
    # - alphas are given in decreasing values
    # - the sum of v's should add up to unity
    if not (alpha1 > alpha2 and 0.0 < v1 <= 1.0):
        raise ValueError(
            "curvefit_t_matrix: alpha 1 must have higher value than alpha 2, and v1 must be a positive fraction"
        )
    alpha = np.array([alpha1, alpha2]).reshape(1, 2)
    v = np.array([v1, 1.0 - v1]).reshape(1, 2)
    alpha_vec = alpha * np.ones((log_length, 1))
    v_vec = v * np.ones((log_length, 1))

    try:
        vp, vsv, _, _ = t_matrix_porosity_c_alpha_v(
            k_min=k_min,
            mu_min=mu_min,
            rho_min=rho_min,
            k_fl=k_fl,
            rho_fl=rho_fl,
            phi=phi,
            perm=perm,
            visco=visco,
            alpha=alpha_vec,
            v=v_vec,
            tau=tau,
            frequency=freq,
            angle=angle_sym_plane,
            frac_inc_con=frac_con,
            frac_inc_ani=frac_ani,
        )
    except ValueError:
        vp = np.zeros(k_min.shape)
        vsv = np.zeros(k_min.shape)

    return np.stack((vp, def_vp_vs_ratio * vsv), axis=1).flatten("F")



import numpy as np
import numpy.typing as npt

from .constant_cement_models import constant_cement_model
from .friable_models import friable_model
from .patchy_cement_model import patchy_cement_model_weight


def curvefit_patchy_cement(
    x_data: npt.NDArray[np.float64],
    weight_k: float,
    weight_mu: float,
    shear_red: float,
    frac_cem: float,
) -> npt.NDArray[np.float64]:
    """Run patchy_cement_model_weight with parameter optimisation for closest possible fit to observations

    Inputs: vp, vs, rho, por, k_fl, rho_fl, p_eff, k_min, mu_min, rho_min, k_cem, mu_cem, rho_cem, phi_c
    Parameters to optimise for: frac_cem, shear_red, weight_k, weight_mu
    Optimise for vp, vs
    """
    # Unpack x inputs
    # In calling function:
    k_min = x_data[:, 0]
    mu_min = x_data[:, 1]
    rho_min = x_data[:, 2]
    k_cem = x_data[:, 3]
    mu_cem = x_data[:, 4]
    rho_cem = x_data[:, 5]
    k_fl = x_data[:, 6]
    rho_fl = x_data[:, 7]
    phi = x_data[:, 8]
    p_eff = x_data[:, 9]
    def_vp_vs_ratio = x_data[0, 10]
    phi_c = x_data[0, 11]

    # Catch phi values that are above phi_c - frac_cem, reduce silently to phi_c - frac_cem
    phi = np.minimum(phi, phi_c - frac_cem)

    try:
        vp, vs, _, _, _ = patchy_cement_model_weight(
            k_min=k_min,
            mu_min=mu_min,
            rho_min=rho_min,
            k_cem=k_cem,
            mu_cem=mu_cem,
            rho_cem=rho_cem,
            k_fl=k_fl,
            rho_fl=rho_fl,
            phi=phi,
            p_eff=p_eff,
            frac_cem=frac_cem,
            phi_c=phi_c,
            coord_num_func="PorBased",
            n=9.0,
            shear_red=shear_red,
            weight_k=weight_k,
            weight_mu=weight_mu,
        )
    except ValueError:
        vp = np.zeros(k_min.shape)
        vs = np.zeros(k_min.shape)

    return np.stack((vp, def_vp_vs_ratio * vs), axis=1).flatten("F")


def curvefit_friable(
    x_data: npt.NDArray[np.float64],
    phi_c: float,
    shear_red: float,
) -> npt.NDArray[np.float64]:
    """Run friable sand model with parameter optimisation for closest possible fit to observations

    Inputs: vp, vs, por, k_fl, rho_fl, p_eff, k_min, mu_min, rho_min
    Parameters to optimise for: phi_c, shear_red
    Optimise for vp, vs
    """
    # Unpack x inputs
    k_min = x_data[:, 0]
    mu_min = x_data[:, 1]
    rho_min = x_data[:, 2]
    k_fl = x_data[:, 3]
    rho_fl = x_data[:, 4]
    phi = x_data[:, 5]
    p_eff = x_data[:, 6]
    def_vp_vs_ratio = x_data[0, 7]

    # Catch phi values that are above phi_c, reduce silently to phi_c
    phi = np.minimum(phi, phi_c)

    try:
        vp, vs, _, _, _ = friable_model(
            k_min=k_min,
            mu_min=mu_min,
            rho_min=rho_min,
            k_fl=k_fl,
            rho_fl=rho_fl,
            phi=phi,
            p_eff=p_eff,
            phi_c=phi_c,
            coord_num_func="PorBased",
            n=1.0,
            shear_red=shear_red,
        )
    except ValueError:
        vp = np.zeros(k_min.shape)
        vs = np.zeros(k_min.shape)

    return np.stack((vp, def_vp_vs_ratio * vs), axis=1).flatten("F")


def curvefit_constant_cement(
    x_data: npt.NDArray[np.float64],
    phi_c: float,
    shear_red: float,
    frac_cem: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """Run constant_cement_model with parameter optimisation for closest possible fit to observations

    Inputs: vp, vs, por, k_min, mu_min, rho_min, k_cem, mu_cem, rho_cem, k_fl, rho_fl
    Parameters to optimise for: phi_c, shear_red, frac_cem
    Optimise for vp, vs
    """
    # Unpack x inputs
    k_min = x_data[:, 0]
    mu_min = x_data[:, 1]
    rho_min = x_data[:, 2]
    k_cem = x_data[:, 3]
    mu_cem = x_data[:, 4]
    rho_cem = x_data[:, 5]
    k_fl = x_data[:, 6]
    rho_fl = x_data[:, 7]
    phi = x_data[:, 8]
    def_vp_vs_ratio = x_data[0, 9]

    # Catch phi values that are above phi_c - frac_cem, reduce silently to phi_c - frac_cem
    phi = np.minimum(phi, phi_c - frac_cem)

    try:
        vp, vs, _, _, _ = constant_cement_model(
            k_min=k_min,
            mu_min=mu_min,
            rho_min=rho_min,
            k_cem=k_cem,
            mu_cem=mu_cem,
            rho_cem=rho_cem,
            k_fl=k_fl,
            rho_fl=rho_fl,
            phi=phi,
            frac_cem=frac_cem,
            phi_c=phi_c,
            n=9.0,
            shear_red=shear_red,
        )
    except ValueError:
        vp = np.zeros(k_min.shape)
        vs = np.zeros(k_min.shape)

    return np.stack((vp, def_vp_vs_ratio * vs), axis=1).flatten("F")



from typing import cast

import numpy as np
import numpy.typing as npt
from matplotlib import pyplot as plt
from matplotlib.colorbar import Colorbar
from matplotlib.text import Text

from .ternary_plot_utilities import (
    make_mesh,
    set_ternary_figure,
    ternary_coord_trans,
)


def ternary_plot(
    data1: npt.NDArray[np.float64],
    data2: npt.NDArray[np.float64],
    data3: npt.NDArray[np.float64],
    data4: npt.NDArray[np.float64],
    well_name: str,
    name_data1: str,
    name_data2: str,
    name_data3: str,
    name_data4: str,
    draw_figures: bool = True,
) -> None:
    """Plot three mineral phases in a ternary plot and use a fourth phase for colour coding."""
    _, ax = set_ternary_figure(
        delta_x=2,
        delta_y=2,
        title="Ternary Plot of Mineral Phases",
        well_name=well_name,
    )

    # Draw background
    vertices = np.array(
        [
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 0.1],
            [1.0, 0.0, 0.0],
        ]
    )
    vertices_xy = ternary_coord_trans(vertices)

    _ = plt.fill(vertices_xy[:, 0], vertices_xy[:, 1], color="w")
    _ = plt.plot(vertices_xy[:, 0], vertices_xy[:, 1], "k", linewidth=1)

    _, _ = make_mesh(ax)

    # Annotate the axes
    label_handles: list[Text] = []
    label_handles.append(
        ax.text(
            x=0.15,
            y=np.sqrt(3) / 4 + 0.05,
            s=name_data1,
            horizontalalignment="center",
            rotation=60,
        )
    )
    label_handles.append(
        ax.text(
            x=0.5,
            y=-0.075,
            s=name_data2,
            horizontalalignment="center",
        )
    )
    label_handles.append(
        ax.text(
            x=0.85,
            y=np.sqrt(3) / 4 + 0.05,
            s=name_data3,
            horizontalalignment="center",
            rotation=-60,
        )
    )
    plt.setp(label_handles, fontname="sans-serif", fontweight="bold", fontsize=11)

    # Plot data
    data_xy = ternary_coord_trans(data1, data2, data3)
    _ = plt.scatter(data_xy[:, 0], data_xy[:, 1], s=64, c=data4, zorder=4)

    hcb = cast(  # Casting due to incomplete type hints in matplotlib
        Colorbar,
        plt.colorbar(),
    )
    hcb.set_label(name_data4, fontname="sans-serif", fontweight="bold", fontsize=11)

    if draw_figures:
        plt.draw()



from typing import Callable, cast

import numpy as np
import numpy.typing as npt
from numpy import float64

from rock_physics_open.equinor_utilities import gen_utilities
from rock_physics_open.equinor_utilities.optimisation_utilities import (
    gen_opt_routine,
    opt_param_to_ascii,
    save_opt_params,
)

from .curvefit_sandstone_models import curvefit_constant_cement


def constant_cement_model_optimisation(
    k_min: npt.NDArray[np.float64],
    mu_min: npt.NDArray[np.float64],
    rho_min: npt.NDArray[np.float64],
    k_cem: npt.NDArray[np.float64],
    mu_cem: npt.NDArray[np.float64],
    rho_cem: npt.NDArray[np.float64],
    k_fl: npt.NDArray[np.float64],
    rho_fl: npt.NDArray[np.float64],
    por: npt.NDArray[np.float64],
    vp: npt.NDArray[np.float64],
    vs: npt.NDArray[np.float64],
    rhob: npt.NDArray[np.float64],
    file_out_str: str = "constant_cement_optimal_params.pkl",
    display_results: bool = False,
    well_name: str = "Unknown well",
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
]:
    """Patchy cement model with optimisation for a selection of parameters.

    Parameters
    ----------
    k_min : np.ndarray
        Cement bulk modulus [Pa].
    mu_min : np.ndarray
        Cement shear modulus [Pa].
    rho_min : np.ndarray
        Cement density [kg/m^3].
    k_cem : np.ndarray
        Cement bulk modulus [Pa].
    mu_cem : np.ndarray
        Cement shear modulus [Pa].
    rho_cem : np.ndarray
        Cement density [kg/m^3].
    k_fl : np.ndarray
        Fluid bulk modulus [Pa].
    rho_fl : np.ndarray
        Fluid density [kg/m^3].
    por : np.ndarray
        Inclusion porosity [ratio].
    vp : np.ndarray
        Compressional velocity log [m/s].
    vs : np.ndarray
        Shear velocity log [m/s].
    rhob : np.ndarray
        Bulk density log [kg/m^3].
    file_out_str : str
        Output file name (string) to store optimal parameters (pickle format).
    display_results : bool
        Display optimal parameters in a window after run.
    well_name : str
        Name of well to be displayed in info box title.

    Returns
    -------
    tuple
        vp_mod, vs_mod - modelled logs, vp_res, vs_res : np.ndarray
        residual logs.
    """

    # Skip hardcoded Vp/Vs ratio
    def_vpvs = np.mean(vp / vs)
    # Set weight to vs to give vp and vs similar influence on optimisation
    y_data = np.stack([vp, vs * def_vpvs], axis=1)
    # Optimisation function for selected parameters
    opt_fun: Callable[..., npt.NDArray[float64]] = curvefit_constant_cement
    # expand single value parameters to match logs length
    por, def_vpvs = cast(
        list[npt.NDArray[np.float64]],
        gen_utilities.dim_check_vector((por, def_vpvs)),
    )
    x_data = np.stack(
        (k_min, mu_min, rho_min, k_cem, mu_cem, rho_cem, k_fl, rho_fl, por, def_vpvs),
        axis=1,
    )

    # Params: weight_k, weight_mu, shear_red, frac_cem
    lower_bound = np.array(
        [
            0.35,  # phi_c
            0.0,  # shear_red
            0.01,  # frac_cem
        ],
        dtype=float,
    )
    upper_bound = np.array(
        [
            0.45,  # phi_c
            1.0,  # shear_red
            0.1,  # frac_cem
        ],
        dtype=float,
    )

    x0 = (upper_bound + lower_bound) / 2.0
    # Optimisation step without fluid substitution
    vel_mod, vel_res, opt_params = gen_opt_routine(
        opt_function=opt_fun,
        x_data_orig=x_data,
        y_data=y_data,
        x_init=x0,
        low_bound=lower_bound,
        high_bound=upper_bound,
    )
    frac_cem = opt_params[2]

    # Reshape outputs and remove weight from vs
    vp_mod, vs_mod = [arr.flatten() for arr in np.split(vel_mod, 2, axis=1)]
    vp_res, vs_res = [arr.flatten() for arr in np.split(vel_res, 2, axis=1)]
    vs_mod = vs_mod / def_vpvs
    vs_res = vs_res / def_vpvs
    vpvs_mod = vp_mod / vs_mod
    # Calculate the modelled density
    rhob_mod = rho_min * (1.0 - por - frac_cem) + frac_cem * rho_cem + por * rho_fl
    ai_mod = vp_mod * rhob_mod
    rhob_res = rhob_mod - rhob
    # Save the optimal parameters
    save_opt_params(
        opt_type="const_cem",
        opt_params=opt_params,
        file_name=file_out_str,
        well_name=well_name,
    )
    if display_results:
        opt_param_to_ascii(in_file=file_out_str, well_name=well_name)

    return vp_mod, vs_mod, rhob_mod, ai_mod, vpvs_mod, vp_res, vs_res, rhob_res



import numpy as np
import numpy.typing as npt


def hertz_mindlin(
    k: npt.NDArray[np.float64],
    mu: npt.NDArray[np.float64],
    phi_c: npt.NDArray[np.float64],
    p: npt.NDArray[np.float64],
    shear_red: npt.NDArray[np.float64] | float,
    coord: npt.NDArray[np.float64] | None = None,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Hertz-Mindlin Pressure-induced moduli increase at critical porosity.

    Parameters
    ----------
    k : np.ndarray
        Bulk modulus of grain mineral [Pa].
    mu : np.ndarray
        Shear modulus of grain mineral [Pa].
    phi_c : np.ndarray
        Critical porosity [fraction].
    p : np.ndarray
        Effective pressure = lithostatic pressure - hydrostatic pressure [Pa].
    shear_red : float or np.ndarray
        Reduced shear factor, if set to 1.0, calculation reduces to standard Hertz-Mindlin equation.
    coord
        coordination number, i.e. the number of grain contacts per grain. If not provided a porosity based
        estimate is used.

    Returns
    -------
    tuple
        k_dry, mu_dry : np.ndarray.
        k_dry:	Bulk modulus [Pa] at effective pressure p,
        mu_dry:	Shear modulus [Pa] at effective pressure p.
    """
    n = 25.98805 * phi_c**2 - 43.7622 * phi_c + 21.6719 if coord is None else coord

    poiss = (3 * k - 2 * mu) / (2 * (3 * k + mu))

    a = ((3 * np.pi * (1 - poiss) * p) / (2 * n * (1 - phi_c) * mu)) ** (1 / 3)

    s_n = (4 * a * mu) / (1 - poiss)
    s_t = (8 * a * mu) / (2 - poiss)

    k_dry = (n * (1 - phi_c) / (12 * np.pi)) * s_n
    mu_dry = (n * (1 - phi_c) / (20 * np.pi)) * (s_n + 1.5 * s_t * shear_red)

    return k_dry, mu_dry



import os

import numpy as np

from rock_physics_open.equinor_utilities.snapshot_test_utilities import (
    INITIATE,
    compare_snapshots,
    get_snapshot_name,
    read_snapshot,
    store_snapshot,
)
from rock_physics_open.t_matrix_models import run_t_matrix, t_matrix_porosity_vectorised

k_min = np.ones(11) * 71.0e9
mu_min = np.ones(11) * 32.0e9
rho_min = np.ones(11) * 2710.0
k_fl = np.ones(11) * 2.7e9
rho_fl = np.ones(11) * 1005.0
phi = np.linspace(0.15, 0.35, 11)
perm = 100.0
visco = 100.0
alpha = np.array([0.8, 0.1, 0.01])
v = np.array([0.89, 0.1, 0.01])
tau = 1e-7 * np.ones_like(alpha)
frequency = 1000
angle = 90.0
frac_inc_con = 0.5
frac_inc_ani = 0.5
pressure = np.array([18.0e6, 28.0e6])


def test_run_t_matrix():
    args = run_t_matrix(
        k_min=k_min,
        mu_min=mu_min,
        rho_min=rho_min,
        k_fl=k_fl,
        rho_fl=rho_fl,
        phi=phi,
        perm=perm,
        visco=visco,
        alpha=alpha,
        v=v,
        tau=tau,
        frequency=frequency,
        angle=angle,
        frac_inc_con=frac_inc_con,
        frac_inc_ani=frac_inc_ani,
        pressure=pressure,
    )

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(tuple(args), read_snapshot(get_snapshot_name()))


def test_run_t_matrix_porosity_vectorised():
    args = t_matrix_porosity_vectorised(
        k_min=k_min,
        mu_min=mu_min,
        rho_min=rho_min,
        k_fl=k_fl,
        rho_fl=rho_fl,
        phi=phi,
        perm=perm * np.ones_like(phi),
        visco=visco * np.ones_like(phi),
        alpha=alpha,
        v=v,
        tau=tau,
        frequency=frequency,
        angle=angle,
        frac_inc_con=frac_inc_con,
        frac_inc_ani=frac_inc_ani,
        pressure=pressure,
    )

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))



from .backus_ave import backus_average
from .dvorkin_nur import dvorkin_contact_cement
from .gassmann import gassmann, gassmann2, gassmann_dry
from .hashin_shtrikman import (
    hashin_shtrikman,
    hashin_shtrikman_average,
    hashin_shtrikman_walpole,
    multi_hashin_shtrikman,
)
from .hertz_mindlin import hertz_mindlin
from .moduli_velocity import moduli, velocity
from .reflection_eq import aki_richards, smith_gidlow
from .rho import rho_b, rho_m
from .voigt_reuss_hill import multi_voigt_reuss_hill, reuss, voigt, voigt_reuss_hill
from .walton import walton_smooth
from .wood_brie import brie, multi_wood, wood

__all__ = [
    "backus_average",
    "dvorkin_contact_cement",
    "gassmann",
    "gassmann2",
    "gassmann_dry",
    "hashin_shtrikman",
    "hashin_shtrikman_average",
    "hashin_shtrikman_walpole",
    "multi_hashin_shtrikman",
    "hertz_mindlin",
    "moduli",
    "velocity",
    "aki_richards",
    "smith_gidlow",
    "rho_b",
    "rho_m",
    "multi_voigt_reuss_hill",
    "reuss",
    "voigt",
    "voigt_reuss_hill",
    "walton_smooth",
    "brie",
    "multi_wood",
    "wood",
]



import pickle
from abc import ABC, abstractmethod
from typing import Any, Self

import numpy as np


class BasePressureModel(ABC):
    """
    Abstract base class for pressure sensitivity models.

    All pressure models follow the convention:
    - predict(): returns differential change (depleted - in_situ)
    - predict_abs(): returns absolute values for specified case
    - predict_max(): uses model_max_pressure instead of depleted pressure

    Input validation is delegated to concrete implementations since
    each model has different column requirements.
    """

    def __init__(self, model_max_pressure: float | None = None, description: str = ""):
        """
        Initialize base pressure model.

        Parameters
        ----------
        model_max_pressure : float | None
            Maximum pressure for predict_max method. Required for predict_max to work.
        description : str
            Human-readable description of the model instance.
        """
        self._model_max_pressure: float | None = model_max_pressure
        self._description: str = description

    @property
    def max_pressure(self) -> float | None:
        """Maximum pressure setting for predict_max method."""
        return self._model_max_pressure

    @property
    def description(self) -> str:
        """Model description."""
        return self._description

    def predict(self, inp_arr: np.ndarray) -> np.ndarray:
        """
        Predict differential change: result(depleted) - result(in_situ).

        Parameters
        ----------
        inp_arr : np.ndarray
            Input array with pressure columns and other model-specific parameters.

        Returns
        -------
        np.ndarray
            Differential change values.
        """
        arr = self.validate_input(inp_arr)
        return self.predict_abs(arr, case="depleted") - self.predict_abs(
            arr, case="in_situ"
        )

    def predict_max(self, inp_arr: np.ndarray) -> np.ndarray:
        """
        Predict using model_max_pressure instead of depleted pressure.

        Parameters
        ----------
        inp_arr : np.ndarray
            Input array where last column (depleted pressure) will be replaced.

        Returns
        -------
        np.ndarray
            Values at model_max_pressure minus values at in_situ pressure.

        Raises
        ------
        ValueError
            If model_max_pressure is not set.
        """
        if self._model_max_pressure is None:
            raise ValueError('Field "model_max_pressure" is not set')

        arr = self.validate_input(inp_arr).copy()
        # Replace last column (assumed to be depleted pressure) with max pressure
        arr[:, -1] = self._model_max_pressure
        return self.predict_abs(arr, case="depleted") - self.predict_abs(
            arr, case="in_situ"
        )

    @abstractmethod
    def validate_input(self, inp_arr: np.ndarray) -> np.ndarray:
        """
        Validate input array format for this specific model.

        Parameters
        ----------
        inp_arr : np.ndarray
            Input array to validate.

        Returns
        -------
        np.ndarray
            Validated input array.

        Raises
        ------
        ValueError
            If input format is invalid for this model.
        """

    @abstractmethod
    def predict_abs(self, inp_arr: np.ndarray, case: str = "in_situ") -> np.ndarray:
        """
        Predict absolute values for specified pressure case.

        Parameters
        ----------
        inp_arr : np.ndarray
            Validated input array.
        case : str
            Either "in_situ" or "depleted" to specify which pressure to use.

        Returns
        -------
        np.ndarray
            Absolute predicted values.
        """

    @abstractmethod
    def todict(self) -> dict[str, Any]:
        """
        Convert model to dictionary for serialization.

        Returns
        -------
        dict[str, Any]
            Dictionary containing all model parameters.
        """

    def save(self, file: str | bytes) -> None:
        """
        Save model to pickle file.

        Parameters
        ----------
        file : str | bytes
            File path for saving.
        """
        with open(file, "wb") as f_out:
            pickle.dump(self.todict(), f_out)

    @classmethod
    @abstractmethod
    def load(cls, file: str | bytes) -> Self:
        """
        Load model from pickle file.

        Parameters
        ----------
        file : str | bytes
            File path for loading.

        Returns
        -------
        BasePressureModel
            Loaded model instance.
        """



import os

import numpy as np

from rock_physics_open.equinor_utilities.snapshot_test_utilities import (
    INITIATE,
    compare_snapshots,
    get_snapshot_name,
    read_snapshot,
    store_snapshot,
)
from rock_physics_open.fluid_models.brine_model import (
    brine_properties,
)
from rock_physics_open.fluid_models.brine_model.brine_properties import (
    brine_density,
    brine_primary_velocity,
    water,
    water_density,
    water_primary_velocity,
)

pres = 23.0e6 * np.linspace(0.8, 1.2, 101)
temp = 100.0 * np.linspace(0.8, 1.2, 101)
sal = 35000.0 * np.ones(101)
nacl = 80.0 * np.ones(101)
cacl = 15.0 * np.ones(101)
kcl = 5.0 * np.ones(101)
gor_co2 = 50 * np.ones(101)


def test_brine_properties():
    args = brine_properties(temp, pres, sal, nacl, kcl, cacl)

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_water_properties():
    args = water(temp, pres)

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_water_density():
    args = water_density(temp, pres)

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_water_velocity():
    args = water_primary_velocity(temp, pres)

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_brine_density():
    args = brine_density(temp, pres, sal)

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_brine_velocity():
    args = brine_primary_velocity(temp, pres, sal)

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_water_brine_properties_float():
    """
    Make sure that input object type is reflected in output type
    """
    for results in [
        water(temp[0], pres[0]),
        water_density(temp[0], pres[0]),
        water_primary_velocity(temp[0], pres[0]),
        brine_properties(temp[0], pres[0], sal[0]),
        brine_density(temp[0], pres[0], sal[0]),
        brine_primary_velocity(temp[0], pres[0], sal[0]),
    ]:
        if hasattr(results, "__iter__"):
            assert all(isinstance(arg, float) for arg in results)
        else:
            assert isinstance(results, float)



from typing import cast

import numpy as np
import numpy.typing as npt

from rock_physics_open.equinor_utilities.gen_utilities import dim_check_vector
from rock_physics_open.equinor_utilities.optimisation_utilities import opt_param_info
from rock_physics_open.equinor_utilities.std_functions import hashin_shtrikman_average
from rock_physics_open.equinor_utilities.various_utilities.types import Array1D, Array2D

from .t_matrix_C import t_matrix_porosity_c_alpha_v


def curvefit_t_matrix_exp(
    x_data: Array2D[np.float64],
    frac_ani: float,
    frac_con: float,
    alpha1: float,
    alpha2: float,
    v1: float,
    k_c: float,
    mu_c: float,
    rho_c: float,
    k_sh: float,
    mu_sh: float,
    rho_sh: float,
) -> Array1D[np.float64]:
    """Optimisation of input parameters to T-Matrix for carbonate in case where the mineral composition for each
    sample is not known, so that effective mineral moduli for each sample are part of the parameters to be optimised
    for. The optimisation is also made for the inclusion parameters in addition to fraction of connected and anisotropic
    porosity and the VTI plane angle.

    Parameters
    ----------
    x_data : np.ndarray
        Inputs to unpack.
    frac_ani : float
        Fraction of anisotropic inclusions.
    frac_con : float
        Fraction of connected inclusions.
    alpha1 : float
        Aspect ratio of first inclusion set.
    alpha2 : float
        Aspect ratio of second inclusion set.
    v1 : float
        Concentration ratio of first inclusion set.
    k_c : float
        p-wave velocity for carbonate matrix.
    mu_c : float
        vp/vs ratio for carbonate matrix.
    rho_c : float
        vp/vs ratio for carbonate matrix.
    k_sh : float
        p-wave velocity for shale.
    mu_sh : float
        vp/vs ratio for shale.
    rho_sh : float
        Density for shale.

    Returns
    -------
    np.ndarray
        Modelled velocity, vp and vs.
    """
    # Restore original value range for parameters - must match the scaling performed in calling function
    _, scale_val, _ = opt_param_info()
    k_c *= scale_val["k_carb"]
    mu_c *= scale_val["mu_carb"]
    rho_c *= scale_val["rho_carb"]
    k_sh *= scale_val["k_sh"]
    mu_sh *= scale_val["mu_sh"]
    rho_sh *= scale_val["rho_sh"]

    # Unpack x inputs
    # In calling function:
    # x_data = np.stack((por, vsh, k_fl, rho_fl, k_r, eta_f, tau, freq, def_vpvs), axis=1)
    phi = x_data[:, 0]
    vsh = x_data[:, 1]
    k_fl = x_data[:, 2]
    rho_fl = x_data[:, 3]
    angle_sym_plane = x_data[0, 4]
    perm = x_data[:, 5]
    visco = x_data[:, 6]
    tau = x_data[:, 7]
    # Both frequency and vp/vs ratio has to be in the same format as the others, but only a scalar value is used
    freq = x_data[0, 8]
    def_vp_vs_ratio = x_data[0, 9]

    # Mineral properties
    # Expand elastic properties to vectors of the same length as the x_data inputs
    k_c_, mu_c_, rho_c_, k_sh_, mu_sh_, rho_sh_, _ = cast(
        list[npt.NDArray[np.float64]],
        dim_check_vector(
            (
                k_c,
                mu_c,
                rho_c,
                k_sh,
                mu_sh,
                rho_sh,
                phi,
            )
        ),
    )
    k_min, mu_min = hashin_shtrikman_average(
        k1=k_sh_,
        mu1=mu_sh_,
        k2=k_c_,
        mu2=mu_c_,
        f=vsh,
    )
    rho_min = vsh * rho_sh_ + (1.0 - vsh) * rho_c_

    # Inclusion aspect ratios and concentrations
    log_len = phi.shape[0]
    alpha = np.zeros((log_len, 2))
    v = np.zeros((log_len, 2))
    alpha[:, 0] = alpha1 * np.ones(log_len)
    alpha[:, 1] = alpha2 * np.ones(log_len)
    v[:, 0] = v1 * np.ones(log_len)
    v[:, 1] = (1 - v1) * np.ones(log_len)

    try:
        vp, vsv, _, _ = t_matrix_porosity_c_alpha_v(
            k_min=k_min,
            mu_min=mu_min,
            rho_min=rho_min,
            k_fl=k_fl,
            rho_fl=rho_fl,
            phi=phi,
            perm=perm,
            visco=visco,
            alpha=alpha,
            v=v,
            tau=tau,
            frequency=freq,
            angle=angle_sym_plane,
            frac_inc_con=frac_con,
            frac_inc_ani=frac_ani,
        )
    except ValueError:
        vp = np.zeros(k_min.shape)
        vsv = np.zeros(k_min.shape)

    return np.stack((vp, def_vp_vs_ratio * vsv), axis=1).flatten("F")



import numpy as np
import numpy.typing as npt
from scipy.constants import gas_constant

from rock_physics_open.equinor_utilities.conversions import celsius_to_kelvin

AIR_WEIGHT = 28.8 * 1.0e-3  # kg/mol


def gas_properties(
    temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
    gas_gravity: npt.NDArray[np.float64],
    model: str | None = None,  # pyright: ignore[reportUnusedParameter]
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
]:
    """
    :param gas_gravity: molar mass of gas relative to air molar mas.
    :param pressure: Formation pressure (Pa)
    :param temperature: Temperature (Celsius).
    :param model: for future use
    :return: vel_gas [m/s], den_gas [kg/m^3], k_gas [Pa], eta_gas [cP]
    """

    den_gas = gas_density(
        absolute_temperature=celsius_to_kelvin(temperature),
        pressure=pressure,
        gas_gravity=gas_gravity,
    )

    k_gas = gas_bulk_modulus(
        absolute_temperature=celsius_to_kelvin(temperature),
        pressure=pressure,
        gas_gravity=gas_gravity,
    )

    vel_gas = (k_gas / den_gas) ** 0.5

    eta_gas = lee_gas_viscosity(
        absolute_temperature=celsius_to_kelvin(temperature),
        pressure=pressure,
        gas_gravity=gas_gravity,
    )

    return vel_gas, den_gas, k_gas, eta_gas


def molecular_weight(gas_gravity: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
    """
    calculates molecular weight of a gas from gas gravity.
    :param gas_gravity: molar mass of gas relative to air molar mas.
    :return: The volume of the gas in kg/mol.
    """
    return gas_gravity * AIR_WEIGHT


def molar_volume(
    absolute_temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    calculates molar volume using the ideal gas law.
    :param absolute_temperature: The absolute temperature of the gas in kelvin.
    :param pressure: Confining pressure in Pa.
    :return: The volume of the gas in m^3/mol.
    """

    return gas_constant * absolute_temperature / pressure


def ideal_gas_density(
    absolute_temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
    gas_gravity: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    calculates molar volume using the ideal gas law.
    :param gas_gravity: molar mass of gas relative to air molar mas.
    :param absolute_temperature: The absolute temperature of the gas in kelvin.
    :param pressure: Confining pressure in Pa.
    :return: The density of the gas in kg/m^3
    """
    return molecular_weight(gas_gravity) / molar_volume(absolute_temperature, pressure)


def ideal_gas_primary_velocity(
    absolute_temperature: npt.NDArray[np.float64],
    gas_gravity: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    :param gas_gravity: molar mass of gas relative to air molar mas.
    :param absolute_temperature: The absolute temperature of the gas in kelvin.
    :return: The compressional wave velocity of the gas in m/s.
    """
    return np.sqrt(gas_constant * absolute_temperature / molecular_weight(gas_gravity))


def ideal_gas(
    absolute_temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
    gas_gravity: npt.NDArray[np.float64],
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    :param gas_gravity: molar mass of gas relative to air molar mas.
    :param absolute_temperature: The absolute temperature of the gas in kelvin.
    :param pressure: Confining pressure in Pa.
    :return: ideal_gas_velocity [m/s], ideal_gas_density [kg/m^3],
    """
    ideal_gas_den = ideal_gas_density(
        absolute_temperature=absolute_temperature,
        pressure=pressure,
        gas_gravity=gas_gravity,
    )
    ideal_gas_vel = ideal_gas_primary_velocity(
        absolute_temperature=absolute_temperature,
        gas_gravity=gas_gravity,
    )
    return ideal_gas_vel, ideal_gas_den


def pseudoreduced_temperature(
    absolute_temperature: npt.NDArray[np.float64],
    gas_gravity: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    calculates pseudoreduced temperature, equation 9a from Batzle & Wang [1].

    Uses relationship from

    Thomas, L. K., Hankinson, R. W., and Phillips, K. A., 1970,
    Determination of acoustic velocities for natural gas: J. Petr.
    Tech., 22, 889-892.

    :param gas_gravity: molar mass of gas relative to air molar mas.
    :param absolute_temperature: The absolute temperature of the gas in kelvin.
    :return: Pseudoreduced temperature in kelvin.
    """
    return absolute_temperature / (94.72 + 170.75 * gas_gravity)


def pseudoreduced_pressure(
    pressure: npt.NDArray[np.float64],
    gas_gravity: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    calculates pseudoreduced pressure, equation 9a from Batzle & Wang [1].

    Uses relationship from

    Thomas, L. K., Hankinson, R. W., and Phillips, K. A., 1970,
    Determination of acoustic velocities for natural gas: J. Petr.
    Tech., 22, 889-892.

    :param gas_gravity: molar mass of gas relative to air molar mas.
    :param pressure: Confining pressure in Pa.
    :return: Pseudoreduced pressure in Pa.
    """
    return pressure / (4.892 - 0.4048 * gas_gravity)


def compressibility_factor(
    absolute_temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
    gas_gravity: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    calculates compressibility hydro-carbon gas, equation 10b and 10c from
    Batzle & Wang [1].

    :param gas_gravity: molar mass of gas relative to air molar mas.
    :param absolute_temperature: The absolute temperature of the gas in kelvin.
    :param pressure: Confining pressure in Pa.
    :return: Gas compressibility - unitless
    """
    tpr = pseudoreduced_temperature(
        absolute_temperature=absolute_temperature,
        gas_gravity=gas_gravity,
    )

    # Pseudoreduced pressure has unit MPa in equation
    ppr = pseudoreduced_pressure(pressure, gas_gravity) * 1.0e-6

    return (
        (0.03 + 0.00527 * (3.5 - tpr) ** 3) * ppr
        + 0.642 * tpr
        - 0.007 * tpr**4
        - 0.52
        + 0.109
        * (3.85 - tpr) ** 2
        / np.exp((0.45 + 8.0 * (0.56 - 1 / tpr) ** 2) * ppr**1.2 / tpr)
    )


def gas_density(
    absolute_temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
    gas_gravity: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    The density of hydro-carbon gas, using equation 10 from Batzle & Wang [1].

    :param gas_gravity: molar mass of gas relative to air molar mas.
    :param absolute_temperature: The absolute temperature of the gas in kelvin.
    :param pressure: Confining pressure in Pa.
    :return: The density of the gas in kg/m^3
    """

    _, ideal_gas_den = ideal_gas(
        absolute_temperature=absolute_temperature,
        pressure=pressure,
        gas_gravity=gas_gravity,
    )
    return ideal_gas_den / compressibility_factor(
        absolute_temperature=absolute_temperature,
        pressure=pressure,
        gas_gravity=gas_gravity,
    )


def compressibility_rate_per_pseudoreduced_pressure(
    absolute_temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
    gas_gravity: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Derivate of compressibility_factor with respect to pressure.

    :param gas_gravity: molar mass of gas relative to air molar mas.
    :param absolute_temperature: The absolute temperature of the gas in kelvin.
    :param pressure: Confining pressure in MPa.
    :return: Derivative of the compressibility factor (unitless) with respect to pseudoreduced pressure
    """
    tpr = pseudoreduced_temperature(absolute_temperature, gas_gravity)

    # Pseudoreduced pressure is expected to be in MPa in the expression
    ppr = pseudoreduced_pressure(pressure, gas_gravity) * 1.0e-6

    return (
        0.03
        + 0.00527 * (3.5 - tpr) ** 3
        - (
            0.1308
            * (0.45 + 8 * (0.56 - tpr ** (-1)) ** 2)
            * (3.85 - tpr) ** 2
            * ppr**0.2
        )
        / (np.exp(((0.45 + 8 * (0.56 - tpr ** (-1)) ** 2) * ppr**1.2) / tpr) * tpr)
    )


def gas_bulk_modulus(
    absolute_temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
    gas_gravity: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    The bulk modulus of hydro-carbon gas, using equation 11 from Batzle & Wang [1].

    :param gas_gravity: molar mass of gas relative to air molar mas.
    :param absolute_temperature: The absolute temperature of the gas in kelvin.
    :param pressure: Confining pressure in Pa.
    :return: The bulk modulus of the gas in Pa.
    """
    z = compressibility_factor(
        absolute_temperature=absolute_temperature,
        pressure=pressure,
        gas_gravity=gas_gravity,
    )
    dz_dppr = compressibility_rate_per_pseudoreduced_pressure(
        absolute_temperature=absolute_temperature,
        pressure=pressure,
        gas_gravity=gas_gravity,
    )

    # Set ppr in unit MPa in order to use it in calculation of gamma_0
    ppr = pseudoreduced_pressure(pressure, gas_gravity) * 1.0e-6

    # Equation 11b
    gamma_0 = (
        0.85
        + 5.6 / (ppr + 2)
        + 27.1 / ((ppr + 3.5) ** 2)
        - 8.7 * np.exp(-0.65 * (ppr + 1))
    )

    return gamma_0 * pressure / (1 - dz_dppr * ppr / z)


def gas_viscosity(
    absolute_temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
    gas_gravity: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    The gas viscosity of hydrocarbon gas, using equations 12 and 13 of Batzle & Wang [1].

    :param absolute_temperature: The absolute temperature of the gas in kelvin.
    :param pressure: Confining pressure in Pa.
    :param gas_gravity: molar mass of gas relative to air mas.
    :return: The gas viscosity of the gas in cP.
    """
    temp_pr = pseudoreduced_temperature(absolute_temperature, gas_gravity)

    # Pseudoreduced pressure should be in unit MPa
    pres_pr = pseudoreduced_pressure(pressure, gas_gravity) * 1.0e-6

    eta_1 = 0.0001 * (
        temp_pr * (28.0 + 48.0 * gas_gravity - 5.0 * gas_gravity**2)
        - 6.47 * gas_gravity**-2
        + 35.0 * gas_gravity**-1
        + 1.14 * gas_gravity
        - 15.55
    )
    return eta_1 * (
        0.001
        * pres_pr
        * (
            (1057.0 - 8.08 * temp_pr) / pres_pr
            + (796.0 * pres_pr**0.5 - 704.0)
            / (((temp_pr - 1.0) ** 0.7) * (pres_pr + 1.0))
            - 3.24 * temp_pr
            - 38.0
        )
    )


def lee_gas_viscosity(
    absolute_temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
    gas_gravity: npt.NDArray[np.float64],
) -> np.ndarray:
    """
    :param absolute_temperature: Absolute temperature of the gas in kelvin.
    :param pressure: Confining pressure in Pa.
    :param gas_gravity: specific gravity of gas relative to air.
    :return: gas viscosity in cP

    Reference
    ---------
    Lee, J. D., et al. (1966). "Viscosity of Natural Gas." In The American Institute of
    Chemical Engineers Journal, Volume 12, Issue 6, pp. 1058-1062.

    Original equation is given in imperial units. Inputs are transformed to temperature
    in Fahrenheit and pressure in psi
    """
    temp_far = (absolute_temperature - 273.15) * 9.0 / 5.0 + 32.0
    pres_psi = pressure / 6894.757
    return (
        0.001
        * (temp_far + 459.67) ** 0.5
        / pres_psi
        * (0.7 + 1.5 * gas_gravity)
        / (gas_gravity + 1) ** 1.5
    )



import numpy as np

from rock_physics_open.equinor_utilities.various_utilities.types import Array3D, Array4D

from .array_functions import array_matrix_mult
from .check_and_tile import check_and_tile


def calc_x_vec(
    s_0: Array3D[np.float64],
    td: Array4D[np.float64],
) -> Array4D[np.float64]:
    """Returns the x-tensor (6x6x(numbers of empty cavities) matrix) for more explanation see e.g.
    Agersborg et al. 2009 or "The effects of drained and undrained loading in
    visco-elastic waves in rock-like composites" M. Jakobsen and T.A. Johansen.
    (2005). Int. J. Solids and Structures (42). p. 1597-1611

    Parameters
    ----------
    s_0: np.ndarray
        Inverse of stiffness tensor of the host material (nx6x6 matrix).
    td :  np.ndarray
        Dry t-matrix tensors (nx6x6x(numbers of empty cavities) matrix).

    Returns
    -------
    np.ndarray
        x-tensor.
    """
    i2_i2, log_length, alpha_length = check_and_tile(s_0, td)

    x = np.zeros((log_length, 6, 6, alpha_length))
    for j in range(alpha_length):
        x[:, :, :, j] = array_matrix_mult(
            td[:, :, :, j], s_0, i2_i2, s_0, td[:, :, :, j]
        )

    return x



import os

import numpy as np

from rock_physics_open.equinor_utilities.snapshot_test_utilities import (
    INITIATE,
    compare_snapshots,
    get_snapshot_name,
    read_snapshot,
    store_snapshot,
)
from rock_physics_open.equinor_utilities.std_functions import moduli
from rock_physics_open.equinor_utilities.various_utilities import (
    gassmann_dry_model,
    gassmann_model,
    gassmann_sub_model,
)
from rock_physics_open.sandstone_models import friable_model, friable_model_dry

k_min = np.ones(11) * 36.8e9
mu_min = np.ones(11) * 44.0e9
rho_min = np.ones(11) * 2650.0
por = np.linspace(0.15, 0.35, 11)
p_eff = np.ones(11) * 10.0e6
k_fl_orig = np.ones(11) * 0.8e9
rho_fl_orig = np.ones(11) * 850.0
k_fl_sub = np.ones(11) * 2.7e9
rho_fl_sub = np.ones(11) * 1005.0

rho_dry = rho_min * (1.0 - por)
k_dry, mu_dry = friable_model_dry(
    k_min=k_min,
    mu_min=mu_min,
    phi=por,
    p_eff=p_eff,
    phi_c=0.4,
    coord_num_func="PorBased",
    n=8,
    shear_red=0.5,
)
vp_sat, vs_sat, rho_sat, _, _ = friable_model(
    k_min=k_min,
    mu_min=mu_min,
    rho_min=rho_min,
    k_fl=k_fl_orig,
    rho_fl=rho_fl_orig,
    phi=por,
    p_eff=p_eff,
    phi_c=0.4,
    coord_num_func="PorBased",
    n=8,
    shear_red=0.5,
)
k_sat, mu_sat = moduli(vp_sat, vs_sat, rho_sat)


def test_gassmann():
    args = gassmann_model(
        k_min=k_min,
        k_fl=k_fl_orig,
        rho_fl=rho_fl_orig,
        k_dry=k_dry,
        mu=mu_dry,
        rho_dry=rho_dry,
        por=por,
    )

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_gassmann_dry():
    args = gassmann_dry_model(
        k_min, k_fl_orig, rho_fl_orig, k_sat, mu_sat, rho_sat, por
    )

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_gassmann_sub():
    args = gassmann_sub_model(
        k_min, k_fl_orig, rho_fl_orig, k_fl_sub, rho_fl_sub, k_sat, mu_dry, rho_sat, por
    )

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))



from typing import Literal, cast

import numpy as np
import numpy.typing as npt

from rock_physics_open.equinor_utilities import std_functions
from rock_physics_open.equinor_utilities.gen_utilities import dim_check_vector

from .dem import dem_model
from .multi_sca import multi_sca


def shale_model_4_mineral_dem(
    k1: npt.NDArray[np.float64],
    mu1: npt.NDArray[np.float64],
    rho1: npt.NDArray[np.float64],
    k2: npt.NDArray[np.float64],
    mu2: npt.NDArray[np.float64],
    rho2: npt.NDArray[np.float64],
    k3: npt.NDArray[np.float64],
    mu3: npt.NDArray[np.float64],
    rho3: npt.NDArray[np.float64],
    k4: npt.NDArray[np.float64],
    mu4: npt.NDArray[np.float64],
    rho4: npt.NDArray[np.float64],
    k_fl: npt.NDArray[np.float64],
    rho_fl: npt.NDArray[np.float64],
    phi: npt.NDArray[np.float64],
    f1: npt.NDArray[np.float64],
    f2: npt.NDArray[np.float64],
    f3: npt.NDArray[np.float64],
    rhob_inp: npt.NDArray[np.float64],
    asp1: npt.NDArray[np.float64],
    asp2: npt.NDArray[np.float64],
    asp3: npt.NDArray[np.float64],
    asp4: npt.NDArray[np.float64],
    asp: npt.NDArray[np.float64],
    mod_type: Literal["SCA", "VRH"] = "SCA",
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
]:
    """
    Simple shale model with mixture of matrix minerals in self-consistent approximation
    or Voigt-Reuss-Hill average. In the latter case the aspect ratios are ignored.
    Fluid filled porosity is included through a DEM inclusion model.

    All k, mu inputs have unit [Pa], all rho inputs have unit [kg/m^3], phi and all f and asp have unit [fraction].

    Parameters
    ----------
    k1, mu1, rho1 : np.ndarray
        Mineral 1 properties [Quartz and feldspar].
    k2, mu2, rho2 : np.ndarray
        Mineral 2 properties [Kerogen].
    k3, mu3, rho3 : np.ndarray
        Mineral 3 properties [Clay].
    k4, mu4, rho4 : np.ndarray
        Mineral 4 properties [Carbonates].
    k_fl, rho_fl : np.ndarray
        Fluid properties.
    phi : np.ndarray
        Phi parameter.
    f1 : np.ndarray
        Fraction of mineral 1.
    f2 : np.ndarray
        Fraction of mineral 2.
    f3 : np.ndarray
        Fraction of mineral 3.
    rhob_inp : np.ndarray
        Observed density [kg/m^3].
    asp1 : np.ndarray
        Aspect ratio mineral 1 inclusions.
    asp2 : np.ndarray
        Aspect ratio mineral 2 inclusions.
    asp3 : np.ndarray
        Aspect ratio mineral 3 inclusions.
    asp4 : np.ndarray
        Aspect ratio mineral 4 inclusions.
    asp : np.ndarray
        Porosity aspect ratio.
    mod_type : str
        One of 'SCA' or 'VRH'.

    Returns
    -------
    tuple
        k, mu, rhob, vp, vs, rho_factor : (np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray).
        k - effective bulk modulus [Pa], mu - effective shear modulus [Pa], rhob - effective density [kg/m^3],
        vp - p-wave velocity [m/s], vs - shear wave velocity [m/s], rho_factor - ratio between modelled and observed
        density [ratio].
    """

    #   tol: DEM model calculation tolerance <= Set as a hardcoded value, not
    #   found to influence the results
    tol = 1.0e-6

    # Calculate effective mineral properties

    # Too harsh to raise exceptions - do a quiet normalisation instead,
    tot = f1 + f2 + f3
    idx = tot > 1.0
    f1[idx] /= tot[idx]
    f2[idx] /= tot[idx]
    f3[idx] /= tot[idx]
    f4 = 1.0 - f1 - f2 - f3

    if mod_type == "SCA":
        k_mat, mu_mat, rho_mat = multi_sca(
            k1,
            mu1,
            rho1,
            f1,
            asp1,
            k2,
            mu2,
            rho2,
            f2,
            asp2,
            k3,
            mu3,
            rho3,
            f3,
            asp3,
            k4,
            mu4,
            rho4,
            f4,
            asp4,
            tol=tol,
        )
    elif mod_type == "VRH":
        k_mat, mu_mat = std_functions.multi_voigt_reuss_hill(
            k1, mu1, f1, k2, mu2, f2, k3, mu3, f3, k4, mu4, f4
        )
        rho_mat = f1 * rho1 + f2 * rho2 + f3 * rho3 + f4 * rho4
    else:
        raise ValueError(
            f'{__file__}: unknown type: {mod_type}, should be one of "SCA", "VRH"'  # pyright: ignore[reportUnreachable] | Kept for backward compatibility
        )

    k_mat, mu_mat, rho_mat, k_fl, rho_fl, phi, asp = cast(
        list[npt.NDArray[np.float64]],
        dim_check_vector((k_mat, mu_mat, rho_mat, k_fl, rho_fl, phi, asp)),
    )
    k, mu, rhob = dem_model(
        k1=k_mat,
        mu1=mu_mat,
        rho1=rho_mat,
        k2=k_fl,
        mu2=np.zeros(len(k_fl)),
        rho2=rho_fl,
        frac2=phi,
        asp2=asp,
        tol=tol,
    )

    rho_factor = rhob / rhob_inp

    vp, vs, _, _ = std_functions.velocity(
        k=k,
        mu=mu,
        rhob=rhob,
    )

    return k, mu, rhob, vp, vs, rho_factor



from typing import TypeVar

import numpy as np

from rock_physics_open.equinor_utilities.various_utilities.types import Array3D

_T = TypeVar("_T", np.float64, np.complex128)


def array_inverse(a: Array3D[_T]) -> Array3D[_T]:
    """Inverse of higher order array (3 dim) using linalg inv routine.

    Parameters
    ----------
    a : np.ndarray
        An nxmxm numpy array.

    Returns
    -------
    np.ndarray
        An nxmxm array where [i, :, :] contains the inverse of A[i, :, :].

    Raises
    ------
    ValueError
        If input of wrong shape.
    """
    if not (a.ndim == 3 and a.shape[1] == a.shape[2]):
        raise ValueError(f"{__name__}: mismatch in inputs variables dimension/shape")
    res = np.empty_like(a)
    res[:] = np.linalg.inv(a)
    return res


def array_matrix_mult(
    *args: Array3D[_T],
) -> Array3D[_T]:
    """3-dim arrays are matrix multiplied args[j][i, :, :] @ args[j+1][i, :, :].
    Input args should all be numpy arrays of shape nxmxm.

    Returns
    -------
    np.ndarray
        An nxmxm array with n args[i] @ args[i+1] @ ....

    Raises
    ------
    ValueError
        If input is not a list or tuple.
    ValueError
        If mismatch in input shape/dimension.
    ValueError
        If mismatch in input shape/dimension.
    """
    if len(args) == 0:
        raise ValueError(f"{__name__}: no input arguments provided")
    if not len(args) > 1:
        return args[0]

    arg_shape: list[tuple[int, ...]] = []
    for i in range(len(args)):
        if not (
            isinstance(args[i], np.ndarray)  # pyright: ignore[reportUnnecessaryIsInstance] | kept for backwards compatibility
            and args[i].ndim == 3
            and args[i].shape[1] == args[i].shape[2]
        ):
            raise ValueError(
                f"{__name__}: mismatch in inputs variables dimension/shape"
            )
        arg_shape.append(args[i].shape)
    if not np.all(np.array(arg_shape)[:, 0] == np.array(arg_shape)[0, 0]):
        raise ValueError(f"{__name__}: mismatch in inputs variables dimension/shape")

    x = args[0]

    for i in range(1, len(args)):
        x = np.einsum("...ij,...jk->...ik", x, args[i])

    return x



import numpy as np

from rock_physics_open.equinor_utilities.various_utilities.types import Array1D, Array3D

from .array_functions import array_matrix_mult


def g_tensor_vec(
    c0: Array3D[np.float64],
    s_0: Array3D[np.float64],
    alpha: Array1D[np.float64] | float,
) -> Array3D[np.float64]:
    """Returns the Eshelby green's tensor (nx6x6 array).

    Parameters
    ----------
    c0 : np.ndarray
        n stiffness tensors of the host material (nx6x6 array).
    s_0 : np.ndarray
        Inverse of stiffness tensor.
    alpha : float | np.ndarray
        Aspect ratio for single inclusion.

    Returns
    -------
    np.ndarray
        g-tensor.

    Raises
    ------
    ValueError
        If mismatch in input dimension/shape.
    """
    if not (
        c0.ndim == 3
        and c0.shape[1] == c0.shape[2]
        and s_0.ndim == 3
        and s_0.shape[1] == s_0.shape[2]
    ):
        raise ValueError(f"{__name__}: mismatch in inputs variables dimension/shape")

    log_len = c0.shape[0]
    mu = c0[:, 3, 3] / 2
    kappa = c0[:, 0, 0] - (4 / 3) * mu
    pois_ratio = (3 * kappa - 2 * mu) / (2 * (3 * kappa + mu))

    s_r = np.zeros(c0.shape)

    s_11 = np.zeros(log_len)
    s_12 = np.zeros(log_len)
    s_13 = np.zeros(log_len)
    s_21 = np.zeros(log_len)
    s_22 = np.zeros(log_len)
    s_23 = np.zeros(log_len)
    s_31 = np.zeros(log_len)
    s_32 = np.zeros(log_len)
    s_33 = np.zeros(log_len)
    s_44 = np.zeros(log_len)
    s_55 = np.zeros(log_len)
    s_66 = np.zeros(log_len)

    if isinstance(alpha, (float, int)):
        alpha = np.ones(log_len) * alpha
    # Check for valid range of alpha - those outside are returned with zero matrices
    idx_in = (alpha >= 0) & (alpha < 1)
    idx_one = alpha == 1
    alpha_in = alpha[idx_in]
    pois_ratio_in = pois_ratio[idx_in]
    pois_ratio_one = pois_ratio[idx_one]

    if np.any(idx_in):
        q = (alpha_in / (1 - alpha_in**2) ** (3 / 2)) * (
            np.arccos(alpha_in) - alpha_in * (1 - alpha_in**2) ** (1 / 2)
        )
        s_11[idx_in] = (3 / (8 * (1 - pois_ratio_in))) * (
            alpha_in**2 / (alpha_in**2 - 1)
        ) + (1 / (4 * (1 - pois_ratio_in))) * (
            1 - 2 * pois_ratio_in - 9 / (4 * (alpha_in**2 - 1))
        ) * q
        s_33[idx_in] = (1 / (2 * (1 - pois_ratio_in))) * (
            1
            - 2 * pois_ratio_in
            + (3 * alpha_in**2 - 1) / (alpha_in**2 - 1)
            - (1 - 2 * pois_ratio_in + 3 * alpha_in**2 / (alpha_in**2 - 1)) * q
        )
        s_12[idx_in] = (1 / (4 * (1 - pois_ratio_in))) * (
            alpha_in**2 / (2 * (alpha_in**2 - 1))
            - (1 - 2 * pois_ratio_in + 3 / (4 * (alpha_in**2 - 1))) * q
        )
        s_13[idx_in] = (1 / (2 * (1 - pois_ratio_in))) * (
            -(alpha_in**2) / (alpha_in**2 - 1)
            + 0.5 * (3 * alpha_in**2 / (alpha_in**2 - 1) - (1 - 2 * pois_ratio_in)) * q
        )
        s_31[idx_in] = (1 / (2 * (1 - pois_ratio_in))) * (
            2 * pois_ratio_in
            - 1
            - 1 / (alpha_in**2 - 1)
            + (1 - 2 * pois_ratio_in + 3 / (2 * (alpha_in**2 - 1))) * q
        )
        s_66[idx_in] = (1 / (4 * (1 - pois_ratio_in))) * (
            alpha_in**2 / (2 * (alpha_in**2 - 1))
            + (1 - 2 * pois_ratio_in - 3 / (4 * (alpha_in**2 - 1))) * q
        )
        s_44[idx_in] = (1 / (4 * (1 - pois_ratio_in))) * (
            1
            - 2 * pois_ratio_in
            - (alpha_in**2 + 1) / (alpha_in**2 - 1)
            - 0.5
            * (1 - 2 * pois_ratio_in - (3 * (alpha_in**2 + 1)) / (alpha_in**2 - 1))
            * q
        )
        s_22[idx_in] = s_11[idx_in]
        s_21[idx_in] = s_12[idx_in]
        s_23[idx_in] = s_13[idx_in]
        s_32[idx_in] = s_31[idx_in]
        s_55[idx_in] = s_44[idx_in]
    if np.any(idx_one):
        s_11[idx_one] = (5 * pois_ratio_one - 1) / (15 * (1 - pois_ratio_one)) + (
            2 * (4 - 5 * pois_ratio_one)
        ) / (15 * (1 - pois_ratio_one))
        s_12[idx_one] = (5 * pois_ratio_one - 1) / (15 * (1 - pois_ratio_one))
        s_13[idx_one] = (5 * pois_ratio_one - 1) / (15 * (1 - pois_ratio_one))
        s_31[idx_one] = (5 * pois_ratio_one - 1) / (15 * (1 - pois_ratio_one))
        s_44[idx_one] = (4 - 5 * pois_ratio_one) / (15 * (1 - pois_ratio_one))
        s_22[idx_one] = s_11[idx_one]
        s_33[idx_one] = s_11[idx_one]
        s_21[idx_one] = s_12[idx_one]
        s_32[idx_one] = s_31[idx_one]
        s_23[idx_one] = s_13[idx_one]
        s_55[idx_one] = s_44[idx_one]
        s_66[idx_one] = s_44[idx_one]

    s_r[:, 0, 0] = s_11
    s_r[:, 1, 1] = s_22
    s_r[:, 2, 2] = s_33
    s_r[:, 3, 3] = 2 * s_44
    s_r[:, 4, 4] = 2 * s_55
    s_r[:, 5, 5] = 2 * s_66
    s_r[:, 0, 1] = s_12
    s_r[:, 0, 2] = s_13
    s_r[:, 1, 0] = s_21
    s_r[:, 1, 2] = s_23
    s_r[:, 2, 0] = s_31
    s_r[:, 2, 1] = s_32

    return array_matrix_mult(-s_r, s_0)
